/*
** Module   : LRUCache.java
** Abstract : LRU cache implementation
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20140925 Created initial version.
*/

package com.goldencode.cache;

/**
 * An implementation of a least-recently-used (LRU) cache. A single element (the least recently
 * used) is expired from the cache when capacity for a new element is needed.
 * <p>
 * Registered listeners are notified of each expiry event.
 * <p>
 * <strong>This implementation is not synchronized</strong>. Any access to the methods of this
 * class which modify the internal state of the cache must be synchronized externally.
 * 
 * @param   K
 *          Type of cache element key.
 * @param   V
 *          Type of cache element value.
 */
public class LRUCache<K, V>
extends ExpiryCache<K, V>
{
   /**
    * Constructor.
    * 
    * @param   limit
    *          Maximum number of elements allowed in the cache.
    * 
    * @throws  IllegalArgumentException
    *          if the specified limit is less than one.
    */
   public LRUCache(int limit)
   {
      super(limit);
   }
   
   /**
    * Constructor.
    * 
    * @param   name
    *          Cache name, for debug and logging purposes.
    * @param   limit
    *          Maximum number of elements allowed in the cache.
    * 
    * @throws  IllegalArgumentException
    *          if the specified limit is less than one.
    */
   public LRUCache(String name, int limit)
   {
      super(name, limit);
   }
   
   /**
    * Get the number of elements which will be expired from the cache each time it becomes full,
    * and capacity is required for a new element.
    * 
    * @return  Number of elements to expire on each clearing pass, in this case 1.
    */
   protected int getCleanLimit()
   {
      return 1;
   }
   
   /**
    * Create a node of the appropriate type for the given key and value.
    * 
    * @param   key
    *          Hashable key.
    * @param   value
    *          Cached value.
    * 
    * @return  A sortable node to store in the cache.
    */
   protected Node<K, V> createNode(K key, V value)
   {
      return new NodeImpl<K, V>(key, value);
   }
   
   /**
    * A sortable node containing the hashable key and value of a cache entry. The node's natural
    * order is determined by the recentness of its last access (creation or retrieval). Less
    * recently accessed nodes sort before more recently accessed nodes.
    * 
    * @param <K>
    *        Hashable key
    * @param <V>
    *        Cached value
    */
   private static class NodeImpl<K, V>
   extends Node<K, V>
   {
      /** Next available node identifier */
      private static long nextAccess = 0;
      
      /** Unique counter assigned each time this element is accessed (created or retrieved) */
      private long accessID;
      
      /**
       * Constructor. Sets initial access time to now, and hit count to 0.
       * 
       * @param   key
       *          Hashable key.
       * @param   value
       *          Cached value.
       */
      NodeImpl(K key, V value)
      {
         super(key, value);
      }
      
      /**
       * Compare this node to another. A node less recently accessed sorts before a more recently
       * accessed node.
       * <p>
       * This algorithm generally leaves the least desirable element first in the sorted result.
       */
      public int compareTo(Node<K, V> n)
      {
         // don't catch ClassCastException, because we control the type of node in the cache
         NodeImpl<K, V> that = (NodeImpl<K, V>) n;
         
         return (int) (this.accessID - that.accessID);
      }
      
      /**
       * An implementation of equivalence checking consistent with <code>compareTo(Node)</code>.
       * 
       * @param   o
       *          Object with which to compare this instance.
       * 
       * @return  <code>true</code> if <code>o</code> is equivalent to this instance, else
       *          <code>false</code>.
       */
      public boolean equals(Object o)
      {
    	  if( !(o instanceof NodeImpl)) return false;
         // don't catch ClassCastException, because we control the type of node in the cache
         NodeImpl<K, V> that = (NodeImpl<K, V>) o;
         
         return (this.accessID == that.accessID);
      }
      
      /**
       * Assign the next available (higher) access ID to this node.
       */
      protected void access()
      {
         accessID = nextAccess++;
      }
   }
}
