/*
** Module   : Expression.java
** Abstract : Provides transparent access to compiled expressions.
**
** Copyright (c) 2001-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050505   @21332 Created initial version. This class was
**                           created to make expression compilation
**                           transparent to users of the expr package.
**                           Lazily compiles user expressions as they are
**                           needed. Manages a cache of compiled
**                           expressions to prevent the same expression
**                           (within the same context) from being compiled
**                           more than once.
** 002 ECF 20050604   @21412 Added detailed error diagnostic reporting.
**                           Writes messages to stderr to help isolate the
**                           problematic portion of a failing expression.
**                           Collapse multi-line infix expressions into a
**                           single line at construction; improves
**                           readability of diagnostic messages.
** 003 ECF 20050614   @21496 Minor change to constructor. No whitespace is
**                           added when a line breaks at method invocation
**                           syntax.
** 004 ECF 20050721   @21789 Minor change to execute method. Now reports
**                           current scope in execution report
**                           diagnostics.
** 005 ECF 20050728   @21973 Debug improvement for 004 (@21789). Now
**                           reports entire stack of scopes which apply
**                           to the current expression at the time of
**                           execution, from innermost to outermost. Only
**                           applies to RT_DEBUG mode.
** 006 GES 20070628   @34317 Moved the compiled expression instance cache
**                           to the symbol resolver class where it can be
**                           an instance var. That class' purpose is to
**                           store data across expressions so it is a more
**                           appropriate location. In addition, since the
**                           cache can now be non-static, this can greatly
**                           reduce the working set of a process which
**                           loads more than 1 expression "engine"
**                           sequentially. The previous approach pinned
**                           all expressions, the symbol resolver, all
**                           libraries... in memory forever!
** 007 ECF 20070629   @34349 Minor enhancements. Removed unnecessary
**                           member from CacheKey. Made remaining members
**                           final. Replaced StringBuffer with
**                           StringBuilder.
** 008 ECF 20080730   @39258 Fixed memory leak. Made reference to symbol
**                           resolver a weak reference, so symbol resolver
**                           can be garbage collected when it is no longer
**                           in use. Also implemented RT_NO_ASSIGN and
**                           RT_CLONE behavior. Abstracted all expression
**                           flags out to the ExpressionFlags interface.
** 009 ECF 20130309          Minor performance tweak:  retrieve system property for line separator
**                           once for entire class, rather than each time an instance is created.
*/

package com.goldencode.expr;

import java.lang.ref.*;

/**
 * This class provides transparent access to compiled expressions. An
 * instance of this class is constructed, given a symbol resolver instance
 * and an infix user expression. The user expression is compiled in a
 * just-in-time fashion, such that expressions which are never actually
 * used do not consume the overhead of compilation. The following methods
 * trigger a just-in-time compilation, if the underlying expression has
 * not yet been compiled at the time of invocation:
 * <ul>
 * <li>{@link #execute}
 * <li>{@link #getReturnType}
 * </ul>
 *
 * Once an expression has been compiled, the resulting instance of the custom
 * subclass of {@link CompiledExpression} generated for the expression is
 * cached in the <code>SymbolResolver</code>. Future accesses of the same
 * expression use the cached instance, rather than recompiling and generating
 * a new instance. The following characteristics of an expression are
 * considered when determining whether it has already been compiled:
 * <ul>
 * <li>the exact text of the infix expression;
 * <li>the {@link Scope current scope} of the symbol resolver at the time of
 *     access;
 * <li>the set of {@link #setRuntimeFlags runtime flags} set for the
 *     expression.
 * </ul>
 *
 * If all of the above attributes of an expression match a compiled
 * expression residing in the cache, then the cached, compiled expression is
 * used. Otherwise, the expression is considered new and it is compiled and
 * cached.
 * <p>
 * Several runtime flag constants can be set to alter the behavior of an
 * expression. If multiple flags are to be set, they are bitwise OR'd
 * together. The <code>setFlags</code> method is used to set these flags.
 * <ul>
 * <li><code>RT_NONE</code>: the default; turns off any special runtime
 *     behavior previously set.
 * <li><code>RT_INT_NARROWING</code>: throws an exception at runtime if an
 *     automatic, integral, narrowing conversion (e.g., <code>long</code>
 *     to <code>int</code>) would cause a loss of information, given the
 *     runtime data value to be converted (<b>not currently implemented</b>).
 * <li><code>RT_FP_WIDENING</code>: throws an exception at runtime if an
 *     automatic widening conversion from an integral value to a floating
 *     point value would cause a loss of information. This occurs when a
 *     very large integral number cannot be properly represented in a
 *     floating point representation (<b>not currently implemented</b>).
 * <li><code>RT_FP_ROUNDING</code>: performs floating point rounding to the
 *     number of decimal places specified with the {@link #setPrecision}
 *     method. This rounding occurs <i>after each arithmetic operation</i>
 *     within the expression, not just on the final result. Note that while
 *     more precise for financial calculations, this rounding adds
 *     <b>significant</b> overhead to an expression's execution time. The
 *     default number of decimal places of precision is 6.
 * <li><code>RT_NO_ASSIGN</code>: supports SQL-like where clause syntax with
 *     respect to the <code>=</code> operator.  This operator is interpreted
 *     as an equivalence test (i.e., <code>==</code>) instead of as an
 *     assignment.
 * <li><code>RT_CLONE</code>: if set, a new instance of the compiled
 *     expression will be created for each execution.  Otherwise, the same
 *     instance will be shared across executions.
 * <li><code>RT_ERROR</code>: this flag should be turned on for debugging
 *     purposes only. It causes the compiler to add additional instructions
 *     to the compiled bytecode array for the compiled expression's
 *     {@link CompiledExpression#execute execute} method. These maintain
 *     diagnostic information about the executing expression. An exception
 *     handler in {@link #execute} uses this information to provide a
 *     detailed error message to <code>stderr</code> in the event of an
 *     execution-time exception, which is useful to pinpoint the part of
 *     an expression which is the cause of the problem. During runs of an
 *     expression compiled without this flag, no exception handler is used,
 *     in order to improve performance.
 * <li><code>RT_DEBUG</code>: this flag should be turned on for debugging
 *     purposes only. It provides a superset of the diagnostic features
 *     supported by the <code>RT_ERROR</code> flag. In addition to those
 *     features, it triggers verbose reporting on the compilation of the
 *     expression and on each execution of the expression. Reporting is
 *     sent to <code>stdout</code>, except in the event of an error, in
 *     which case it is sent to <code>stderr</code>, as with
 *     <code>RT_ERROR</code>.
 * </ul>
 */
public final class Expression
implements ExpressionFlags
{
   /** Line separator */
   private static final String LINE_SEP = System.getProperty("line.separator");
   
   /** Symbol resolver used to compile this expression */
   private final WeakReference<SymbolResolver> resolver;
   
   /** Infix (pre-parsed) expression */
   private final String infix;
   
   /** Bitset of runtime safety checks to be performed */
   private int flags = RT_NONE;
   
   /** Compiled expression instance */
   private CompiledExpression compiled = null;
   
   /** Number of decimal places of precision for double rounding */
   private int precision = 6;
   
   /** Flag used to determine whether a cache check is necessary */
   private boolean dirty = false;
   
   /**
    * Construct a new instance of an <code>Expression</code> using the
    * specified symbol resolver and infix notation expression.
    * <p>
    * Any line breaks are stripped from the infix phrase, as is any leading
    * and trailing whitespace from each line.
    *
    * @param   resolver
    *          Object which will define the scope for this expression and
    *          which will be used to resolve symbols encountered within the
    *          expression at parse time into constants, library or variable
    *          references, or method calls.
    * @param   infix
    *          The text of the expression, in infix notation (i.e.,
    *          <code>A = B + C</code>). May not be <code>null</code>.
    */
   public Expression(SymbolResolver resolver, String infix)
   {
      // Collapse multi-line expressions into one line.
      StringBuilder buf = new StringBuilder();
      String[] lines = infix.split(LINE_SEP);
      int len = lines.length;
      for (int i = 0; i < len; i++)
      {
         String next = lines[i].trim();
         if (i > 0)
         {
            int l = buf.length();
            if (l > 0 && buf.charAt(l - 1) != '.' && !next.startsWith("."))
            {
               buf.append(' ');
            }
         }
         buf.append(next);
      }
      
      // Save instance variables.
      this.resolver = new WeakReference<SymbolResolver>(resolver);
      this.infix = buf.toString();
   }
   
   /**
    * Report the type of object which will be returned by the expression
    * when it is {@link #execute executed}. Expressions which would normally
    * return <code>void</code> will report a <code>null</code> return type.
    * Expressions which would normally return a primitive value will report
    * that primitive value's wrapper class. Expressions which return an
    * object reference will report that reference's class or interface.
    * <p>
    * <b>Note:</b> invocation of this method will trigger a lazy compile of
    * the underlying expression, if it has not previously been compiled.
    *
    * @return  Class which represents the expression's return type, or
    *          <code>null</code> if the expression should return
    *          <code>void</code>.
    */
   public Class<?> getReturnType()
   {
      return getCompiledInstance().getReturnType();
   }
   
   /**
    * Report the bitfield of runtime flags currently set for this expression
    * object.
    *
    * @return  Runtime flags.
    */
   public int getRuntimeFlags()
   {
      return flags;
   }
   
   /**
    * Set the bitfield of runtime flags for this expression object.
    *
    * @param   flags
    *          Runtime flags.
    */
   public void setRuntimeFlags(int flags)
   {
      if (this.flags != flags)
      {
         dirty = true;
      }
      this.flags = flags;
   }
   
   /**
    * Report the number of decimal places of precision currently set for
    * floating point rounding operations for this expression object. Note
    * this value is only used if the <code>RT_FP_ROUNDING</code> runtime
    * flag is set; otherwise, it is ignored.
    *
    * @return  Number of decimal places of precision.
    */
   public int getPrecision()
   {
      return precision;
   }
   
   /**
    * Set the number of decimal places of precision to use for floating point
    * rounding operations for this expression object. Note that this value
    * is only used if the <code>RT_FP_ROUNDING</code> runtime flag is set;
    * otherwise, it is ignored.
    *
    * @param   precision
    *          Number of decimal places of precision.
    */
   public void setPrecision(int precision)
   {
      this.precision = precision;
      dirty = (dirty || (flags & RT_FP_ROUNDING) != 0);
   }
   
   /**
    * Execute the expression and return its result. The expression will be
    * compiled and instantiated first, if it does not already exist in the
    * cache.
    * <p>
    * If the <code>RT_DEBUG</code> flag is set, an expression execution debug
    * report is generated to <code>stdout</code>. This report includes the
    * class name of the custom, compiled expression, the original infix
    * expression string, the result generated by the execution (if execution
    * completes successfully), and the state of the variable pool associated
    * with the expression's scope, after execution.
    * <p>
    * If either the <code>RT_ERROR</code> or <code>RT_DEBUG</code> flag is
    * set, the exception handler for this method catches any exception thrown
    * while executing the expression and writes an error message to
    * <code>stderr</code> which provides detail about which part of the
    * expression was executing when the exception occurred. This helps the
    * end user pinpoint the exact problem within a failing expression.
    * <p>
    * <b>This method must be allowed to perform all of its work within the
    * context of a single scope.</b> That is, each and every call to the
    * {@link SymbolResolver#getCurrentScope} method of the symbol resolver
    * associated with this expression made while this method is executing
    * must return the same {@link Scope} object. Otherwise, proper
    * compilation of the expression may be compromised and the results will
    * be undefined.
    *
    * @return  The object which represents the result of executing the
    *          compiled expression. This object will be of the type reported
    *          by {@link #getReturnType}.
    */
   public Object execute()
   {
      if ((flags & (RT_ERROR | RT_DEBUG)) == 0)
      {
         // No exception handling or reporting; just do it.
         return getCompiledInstance().execute();
      }
      
      // Report diagnostics.
      CompiledExpression ce = null;
      try
      {
         ce = getCompiledInstance();
         
         if ((flags & RT_DEBUG) != 0)
         {
            System.out.println("---------------------------");
            System.out.println("EXPRESSION EXECUTION REPORT");
            System.out.println("---------------------------");
            System.out.println("Class :  " + ce.getClass().getName());
            System.out.println("Infix :  " + infix);
         }
         
         Object result = ce.execute();
         
         if ((flags & RT_DEBUG) != 0)
         {
            System.out.println("Result:  " + result);
         }
         
         return result;
      }
      catch (Exception exc)
      {
         // If error occurred before we got the compiled expression instance,
         // we can't do much other than rethrow it.
         if (ce == null)
         {
            if (!(exc instanceof RuntimeException))
            {
               exc = new ExpressionException(
                  "Error creating expression", exc);
            }
            
            throw ((RuntimeException) exc);
         }
         
         // Write details about the error to stderr.
         
         int line = 0;
         int col = 0;
         
         synchronized (ce)
         {
            System.err.println("EXPRESSION EXECUTION ERROR:");
            System.err.println("---------------------------");
            ExtraAst tree = ce.getAst();
            int failingId = ce.getActiveAstId();
            ExtraAst cause = reportFailingNode(tree, failingId, 0, null);
            
            // Compose error message.
            StringBuilder buf = new StringBuilder("{ ");
            String msg = exc.getMessage();
            msg = (msg != null && msg.trim().length() > 0
                   ? msg
                   : exc.getClass().getName());
            buf.append(msg).append(" }");
            msg = buf.toString();
            
            // If we have enough information, print the infix expression
            // with a marker message under the failing location.
            if (cause != null && (line = cause.getLine()) > 0)
            {
               col = cause.getColumn();
               String[] lines = infix.split(LINE_SEP);
               int len = lines.length;
               for (int i = 0; i < len; i++)
               {
                  System.err.println(lines[i]);
                  if (i == line - 1)
                  {
                     buf = new StringBuilder();
                     for (int j = 1; j < col; j++)
                     {
                        buf.append(' ');
                     }
                     buf.append("^  ").append(msg);
                     System.err.println(buf.toString());
                  }
               }
               System.err.println("---------------------------");
            }
            
            // If we didn't have enough information to print the marker
            // message above, or if the user requires verbose debug
            // output, print the postfix AST tree, and mark the failing
            // node.
            if (cause == null || line == 0 || ((flags & RT_DEBUG) != 0))
            {
               reportFailingNode(tree, failingId, 0, msg);
               System.err.println("---------------------------");
            }
            
         }
         
         String msg = "Expression execution error";
         if (line > 0)
         {
            StringBuilder buf = new StringBuilder(msg);
            buf.append(" @")
               .append(line)
               .append(':')
               .append(col);
            msg = buf.toString();
         }
         
         throw new ExpressionException(msg, exc);
      }
      finally
      {
         if ((flags & RT_DEBUG) != 0)
         {
            SymbolResolver sr = resolver.get();
            System.out.print("Scopes:  ");
            Scope scope = sr.getCurrentScope();
            int i;
            for (i = 0; scope != null; i++)
            {
               if (i > 0)
               {
                  System.out.print(", ");
               }
               System.out.print(scope.toString());
               
               scope = scope.getEnclosingScope();
            }
            if (i == 0)
            {
               System.out.print("none!");
            }
            System.out.println();
            sr.dumpVariablePool();
            System.out.println("------- END REPORT --------");
         }
      }
   }
   
   /**
    * A dual-purpose method which both isolates the AST node within an
    * expression's syntax tree which is the cause of a runtime failure, and
    * optionally prints diagnostic information about the faiure to
    * <code>stderr</code>. The optional diagnostic dump is triggered by
    * providing a non-<code>null</code> <code>errorMessage</code> parameter.
    * <p>
    * In the latter mode, the expression's postfix form (an indented AST
    * representation) is written, with <code>errorMessage</code> annotated
    * to the failing node.
    * <p>
    * This method is recursive, implementing a depth-first iteration of the
    * expression's AST. If <code>errorMessage</code> is <code>null</code>,
    * the iteration terminates as soon as the node whose ID is
    * <code>failingID</code> is visited.
    *
    * @param   ast
    *          The postfix (AST) representation of the parsed infix
    *          expression.
    * @param   failingId
    *          Unique identifier associated with the failing node within
    *          the expression's AST.
    * @param   level
    *          Zero-based AST node level currently being visited in this
    *          method's recursive iteration of the tree.
    * @param   errorMessage
    *          An optional error message which will annotate the failing
    *          AST node in a dump of the tree to <code>stderr</code>. If
    *          <code>null</code>, no such dump is written.
    *
    * @return  The AST node whose ID is <code>failingId</code>.
    */
   private ExtraAst reportFailingNode(ExtraAst ast,
                                      int failingId,
                                      int level,
                                      String errorMessage)
   {
      ExtraAst cause = null;
      
      if (ast.getId() == failingId)
      {
         cause = ast;
      }
      
      if (errorMessage != null)
      {
         StringBuilder buf = new StringBuilder();
         for (int i = 0; i < level; i++)
         {
            buf.append("   ");
         }
         
         buf.append("[ ")
            .append(ast.getText())
            .append(" ] <")
            .append(ExpressionParser._tokenNames[ast.getType()])
            .append("> @")
            .append(ast.getLine())
            .append(":")
            .append(ast.getColumn());
            
         if (cause != null)
         {
            buf.append("  <--- ").append(errorMessage);
         }
         
         System.err.println(buf.toString());
      }
      
      ExtraAst child = (ExtraAst) ast.getFirstChild();
      if (child != null)
      {
         level++;
      }
      
      while (child != null)
      {
         ExtraAst tmp = reportFailingNode(child,
                                          failingId,
                                          level,
                                          errorMessage);
         if (tmp != null)
         {
            cause = tmp;
            if (errorMessage == null)
            {
               break;
            }
         }
         
         child = (ExtraAst) child.getNextSibling();
      }
      
      return cause;
   }
   
   /**
    * Report the infix notation expression string provided at construction
    * of this object.
    *
    * @return  Infix expression text.
    */
   public String getInfix()
   {
      return infix;
   }
   
   /**
    * Report the symbol resolver associated with this object at construction.
    *
    * @return  This expression's symbol resolver.
    */
   SymbolResolver getResolver()
   {
      return resolver.get();
   }
   
   /**
    * Get a configured instance of the custom, compiled expression class
    * associated with this expression's current state. This method takes the
    * following steps to return the instance:
    * <ol>
    * <li>if the expression already has been compiled and the compiled class
    *     instantiated, <b>and</b> no changes have been made which would
    *     require a recompile, the stored instance is returned; else
    * <li>the cache is checked for an instance of a compiled expression which
    *     matches the information stored within this object; if one is found,
    *     the cached instance is returned; else
    * <li>the infix expression is compiled and the compiled class is
    *     instantiated; the new instance is cached and returned.
    * </ol>
    *
    * @return  The compiled expression instance associated with this object.
    */
   private CompiledExpression getCompiledInstance()
   {
      // Avoid cache check if we can...
      if (!dirty && compiled != null)
      {
         return compiled;
      }
      
      // Check cache; compile if necessary.
      SymbolResolver sr = resolver.get();
      CacheKey key = new CacheKey(this);
      Scope scope = sr.getCurrentScope();
      sr.lockInstanceCache();
      CompiledExpression instance = sr.getCachedInstance(scope, key);
      if (instance != null)
      {
         sr.unlockInstanceCache();
      }
      
      if (instance == null)
      {
         // Nothing in the cache; we must compile.
         try
         {
            compiled = Compiler.compile(this);
            sr.putCachedInstance(scope, key, compiled);
         }
         catch (CompilerException exc)
         {
            // Wrap this in an unchecked exception, so the caller does not
            // take the mandatory performance hit of a try-catch around
            // every call to execute the expression.
            throw new ExpressionException(
               "Expression error [" + infix + "]", exc);
         }
         finally
         {
            sr.unlockInstanceCache();
         }
      }
      else if ((flags & RT_CLONE) != 0)
      {
         compiled = instance.clone(sr);
      }
      else
      {
         compiled = instance;
      }
      
      dirty = false;
      
      return compiled;
   }
   
   /**
    * Definition of the key to the cache of {@link CompiledExpression}
    * instances. When performing a lookup in the cache, the following
    * expression attributes are considered:
    * <ul>
    * <li>the exact text of the infix expression;
    * <li>the set of {@link #setRuntimeFlags runtime flags} set for the
    *     expression.
    * </ul>
    */
   private static class CacheKey
   {
      /** Expression text in infix notation */
      private final String infix;
      
      /** Runtime flags set for expression */
      private final int flags;
      
      /**
       * Convenience constructor which instantiates a new key and sets the
       * internal key state, based upon the expression state.
       *
       * @param   expr
       *          Expression object on which the cache key state is based.
       */
      CacheKey(Expression expr)
      {
         this.infix = expr.infix;
         this.flags = expr.flags;
      }
      
      /**
       * Generates a hash code which considers the internal state of this
       * key and which is consistent with the implementation of {@link
       * #equals}.
       *
       * @return  Hash code value for this key.
       */
      public int hashCode()
      {
         int result = 17;
         result = 37 * result + flags;
         result = 37 * result + infix.hashCode();
         
         return result;
      }
      
      /**
       * Implementation of <code>equals</code> which considers all relevant
       * state variables in this object.
       *
       * @param   o
       *          Another <code>CacheKey</code> instance to compare against
       *          this one.
       *
       * @return  <code>true</code> if this object is equivalent to
       *          <code>o</code> else <code>false</code>.
       */
      public boolean equals(Object o)
      {
    	  if(!(o instanceof CacheKey)) return false; 
         // Assume case is safe.
         CacheKey that = (CacheKey) o;
         
         return (infix.equals(that.infix) && flags == that.flags);
      }
   }
   
   /**
    * Entry point for manual testing of this class. Uses the simple test
    * symbol resolver provided by {@link SymbolResolver}. It executes an
    * arbitrary expression 1 or more times and reports the results.
    *
    * @param   args
    *          The following arguments are expected:
    *          <ol>
    *          <li>A positive integer representing the number of times to
    *              execute an expression.
    *          <li>An infix notation expression to execute. This expression
    *              can only contain references that are exposed by the
    *              {@link SymbolResolver.TestResolver test resolver}.
    *          <li>[optional] A hex value representing the bitwise OR'd
    *              runtime flags to apply to the expression.
    *          <li>[optional] The number of decimal places of precision to
    *              apply if the previous argument sets the
    *              <code>RT_FP_ROUNDING</code> flag.
    *          </ol>
    */
   public static void main(String[] args)
   {
      try
      {
         int iterations = Integer.parseInt(args[0]);
         SymbolResolver resolver = SymbolResolver.getTestResolver();
         Expression expr = new Expression(resolver, args[1]);
         if (args.length > 2)
         {
            expr.setRuntimeFlags(Integer.parseInt(args[2], 16));
         }
         if (args.length > 3)
         {
            expr.setPrecision(Integer.parseInt(args[3]));
         }
         long start = System.currentTimeMillis();
         Object result = expr.execute();
         for (int i = 1; i < iterations; i++)
         {
            expr.execute();
         }
         long elapsed = System.currentTimeMillis() - start;
         System.out.println("result:  " + result);
         System.out.println(args[0] + " iteration(s) took " + elapsed +
                            " milliseconds");
      }
      catch (Exception exc)
      {
         exc.printStackTrace();
      }
   }
}