/*
** Module   :DateValue.java
** Abstract :Date value holder.
**
** Copyright (c) 2005, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 SIY 20050218 ADD  @20017  Created initial version
** 002 SIY 20050310 CHG  @20356  Organized imports and reordered methods.
**                               Fixed format string.
**                               Fixed month calculation.
**                               Fixed equals();
** 003 SIY 20050323 CHG  @20470  Fixed formatting.
** 004 SIY 20050325 CHG  @20493  Fixed format string.
** 005 SIY 20050328 CHG  @20576  Added "implements Comparable" and cleared 
**                               milliseconds field.
** 006 NVS 20050426 CHG  @20890  Class implements Serializable interface.
** 007 SIY 20050429 CHG  @21009  Fixed formatting.  
*/

package com.goldencode.p2j.directory;

import java.text.DateFormat;
import java.text.FieldPosition;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.io.Serializable;

/**
 * The main purpose of this class is to hold date value for storing/saving it
 * in the Directory. Also it allows to distinguish types during processing
 * (this is not so simple for java.util.Date and derivatives). For the
 * convenience it can be easily converted to/from java.util.Date class.
 * 
 * @author  SIY
 * @version 1.0
 */
public class DateValue
   implements Comparable, Serializable
{
   /**
    * Format used for <code>DateValue</code> <-><code>String</code>
    * conversion.
    */
   private static final ThreadLocal<SimpleDateFormat> format = new ThreadLocal<SimpleDateFormat>() {
       @Override 
       protected SimpleDateFormat initialValue() {
           return new SimpleDateFormat("yyyy-MM-dd");
       }
   };

   /** Internal instance of the Calendar */
   private Calendar calendar;

   /**
    * Construct an instance for the current date.
    */
   public DateValue()
   {
      calendar = Calendar.getInstance();
      enforceTime(calendar.getTime());
   }

   /**
    * Construct new instance from the instance of the <code>Date</code>.
    * 
    * @param   other
    *          An instance which is used as a source of the information.
    */
   public DateValue(Date other)
   {
      calendar = Calendar.getInstance();
      enforceTime(other);
   }

   /**
    * Construct new instance from the other instance of the
    * <code>DateValue</code>.
    * 
    * @param   other
    *          An instance which is used as a source of the information.
    */
   public DateValue(DateValue other)
   {
      calendar = Calendar.getInstance();
      enforceTime(other.getDate());
   }

   /**
    * Construct an instance for specified date.
    * 
    * @param   year
    *          Year value for the new instance.
    * @param   month
    *          Month value for the new instance.
    * @param   day
    *          Day of the month value for the new instance.
    */
   public DateValue(int year, int month, int day)
   {
      calendar = Calendar.getInstance();
      calendar.set(year, month - 1, day, 0, 0, 0);
      enforceTime(calendar.getTime());
   }

   /**
    * Parse <code>String</code> like YYYYMMDD and create an instance of
    * <code>DateValue</code>.
    * 
    * @param   str
    *          Source string.
    * @return  new <code>DateValue</code> instance or <code>null</code> if
    *          source string does not contain properly formatted date.
    */
   public static DateValue valueOf(String str)
   {
      try
      {
         return new DateValue(format.get().parse(str, new ParsePosition(0)));
      }
      catch (Exception e)
      {
         return null;
      }
   }

   /**
    * Compares this instance with instance of <code>Date</code> for
    * ordering.
    * 
    * @param   other
    *          The instance of <code>Date</code> to be compared.
    * @return  <code>0</code> if dates are equal; a value less than
    *          <code>0</code> if this instance date is before date of the
    *          passed parameter; a value greater than <code>0</code> if this
    *          instance date is after date of the parameter.
    */
   public int compareTo(Date other)
   {
      return calendar.getTime().compareTo(other);
   }

   /**
    * Compares two instances for ordering.
    * 
    * @param   other
    *          The instance to be compared.
    * @return  <code>0</code> if instances are equal; a value less than
    *          <code>0</code> if this instance date is before date of the
    *          instance passed as a parameter; a value greater than
    *          <code>0</code> if this instance date is after date of the
    *          instance passed as a parameter.
    */
   public int compareTo(DateValue other)
   {
      return calendar.getTime().compareTo(other.getDate());
   }

   /**
    * Compare this instance with other object.
    * 
    * @param   other
    *          The instance of <code>Date</code> to be compared.
    * @return  <code>0</code> if dates are equal; a value less than
    *          <code>0</code> if this instance date is before date of the
    *          passed parameter; a value greater than <code>0</code> if this
    *          instance date is after date of the parameter.
    * @throws  ClassCastException
    *          if argument is not instance of <code>Date</code> or
    *          <code>DateValue</code>.
    */
   public int compareTo(Object other)
   {
      if (other instanceof Date)
         return calendar.getTime().compareTo((Date)other);

      return compareTo((DateValue)other);
   }

   /**
    * Compares two objects for equality. Calls <code>Calendar.equals()</code>
    * or <code>Date.equals()</code> if possible. Otherwise return
    * <code>false</code>.
    * 
    * @param   o
    *          Object to compare with.
    * @return  <code>true</code> if instances are equal and
    *          <code>false</code> otherwise.
    */
   public boolean equals(Object o)
   {
      if (o instanceof DateValue || o instanceof Calendar)
      {
         Calendar other;

         if (o instanceof DateValue)
            other = ((DateValue)o).calendar;
         else
            other = (Calendar)o;

         if (calendar.get(Calendar.YEAR) == other.get(Calendar.YEAR)
            && calendar.get(Calendar.MONTH) == other.get(Calendar.MONTH)
            && calendar.get(Calendar.DATE) == other.get(Calendar.DATE))
         {
            return true;
         }
         return false;
      }

      if (o instanceof Date)
         return getDate().equals(o);

      return false;
   }

   /**
    * Return current value represented as an instance of <code>Date</code>.
    * 
    * @return  an instance of <code>java.util.Date</code>.
    */
   public Date getDate()
   {
      return calendar.getTime();
   }

   /**
    * Return a hash code for the instance. Use <code>Date.hashCode()</code>
    * for real computation.
    * 
    * @return  hash code for the instance.
    */
   public int hashCode()
   {
      return calendar.getTime().hashCode();
   }

   /**
    * Set current value using instance of <code>Date</code> as a source.
    * 
    * @param   date
    *          Source instance of <code>Date</code> to get values from.
    */
   public void setDate(Date date)
   {
      enforceTime(date);
   }

   /**
    * Convert <code>DateValue</code> instance into <code>String</code>.
    * 
    * @return  string in format YYYYMMMDD.
    */
   public String toString()
   {
      StringBuffer buf = new StringBuffer();

      format.get().format(getDate(), buf, new FieldPosition(DateFormat.YEAR_FIELD));

      return buf.toString();
   }

   /**
    * This method performs all necessary steps to enforce DateValue instance
    * to contain only date, without time part.
    * 
    * @param   input
    *          <code>Date</code> instance to set information from.
    */
   private void enforceTime(Date input)
   {
      //Make sure that time part is set to 0
      calendar.setTime(input);
      calendar.set(Calendar.HOUR_OF_DAY, 0);
      calendar.set(Calendar.MINUTE, 0);
      calendar.set(Calendar.SECOND, 0);
      calendar.set(Calendar.MILLISECOND, 0);
   }
}
