/*
** Module   : HQLHelperCache.java
** Abstract : Context-specific cache for HQLHelper object instances
**
** Copyright (c) 2004-2009, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20051007 ADD   @23249 Created initial version. Caches HQLHelper
**                               objects available in the current context for
**                               faster access to queries.
** 002 ECF 20060120 CHG   @24043 Integrated with P2J server. Replaced
**                               ThreadLocal with ContextLocal.
** 003 ECF 20060306 CHG   @25022 Added support for foreign relation joins. The
**                               foreign DMO interface was added as an
**                               additional, optional, secondary key
**                               component.
** 004 ECF 20070228 CHG   @32245 Added idOnly designation to key. Designates a
**                               projection query returning the primary key ID
**                               only.
** 005 ECF 20070827 CHG   @35001 Minor optimization. Replace StringBuffer with
**                               StringBuilder.
** 006 ECF 20080310 CHG   @37476 Added Hibernate types to key.
** 007 ECF 20090204 CHG   @41256 Improved performance. Replaced string-based
**                               cache key with a faster implementation.
** 008 ECF 20090205 CHG   @41260 Fixed NPE in Key's hashCode() and equals()
**                               methods.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.security.*;

/**
 * A cache of {@link HQLHelper} objects available in the current context,
 * mapped first by record buffer, and within this mapping, by restriction
 * sorting, and foreign key buffer criteria.  The first level mapping is
 * keyed by {@link RecordBuffer} instances via a weak hash map, so these
 * entries are eligible for garbage collection as these buffer instances are
 * discarded.  The second level mapping (within the scope of a particular
 * record buffer) is keyed by a combination of a query's base where clause,
 * base order by clause, and foreign key interface, if any.  Any or all of
 * these secondary criteria may be <code>null</code>.
 * <p>
 * An <code>HQLHelper</code> instance is retrieved using the {@link #get}
 * method;  it is stored in the cache using the {@link #put} method.
 */
final class HQLHelperCache
{
   /** Context local variable containing a map of helper objects by buffer */
   private final ContextLocal<Map<RecordBuffer, Map<Key, HQLHelper>>> local =
      new ContextLocal<Map<RecordBuffer, Map<Key, HQLHelper>>>()
   {
      protected Map<RecordBuffer, Map<Key, HQLHelper>> initialValue()
      {
         return (new WeakHashMap<RecordBuffer, Map<Key, HQLHelper>>());
      }
   };
   
   /**
    * Default constructor.
    */
   HQLHelperCache()
   {
   }
   
   /**
    * Retrieve a query from the cache, based on the specified buffer,
    * where clause, and sort clause.
    *
    * @param   buffer
    *          Record buffer associated with the helper.
    * @param   types
    *          Hibernate types of query substitution parameters.
    * @param   idOnly
    *          Indicates whether query is a projection query returning the
    *          primary key ID only.
    * @param   where
    *          Where clause of a query.
    * @param   sort
    *          Order by clause of a query.
    * @param   joinIface
    *          Interface DMO associated within query by foreign key.  May be
    *          <code>null</code>.
    *
    * @return  A cached <code>HQLHelper</code> instance or <code>null</code>
    *          if an existing instance was not found in the cache.
    */
   HQLHelper get(RecordBuffer buffer,
                 Type[] types,
                 boolean idOnly,
                 String where,
                 String sort,
                 Class<?> joinIface)
   {
      HQLHelper helper = null;
      
      Map<Key, HQLHelper> cached = local.get().get(buffer);
      if (cached != null)
      {
         Key key = new Key(where, types, idOnly, sort, joinIface);
         helper = cached.get(key);
      }
      
      return helper;
   }
   
   /**
    * Store a query in the cache, indexed by the specified buffer,
    * where clause, and sort clause.
    *
    * @param   buffer
    *          Record buffer associated with the helper.
    * @param   types
    *          Hibernate types of query substitution parameters.
    * @param   idOnly
    *          Indicates whether query is a projection query returning the
    *          primary key ID only.
    * @param   where
    *          Where clause of a query.
    * @param   sort
    *          Order by clause of a query.
    * @param   joinIface
    *          Interface DMO associated within query by foreign key.  May be
    *          <code>null</code>.
    * @param   helper
    *          <code>HQLHelper</code> instance to be cached.
    */
   void put(RecordBuffer buffer,
            Type[] types,
            boolean idOnly,
            String where,
            String sort,
            Class<?> joinIface,
            HQLHelper helper)
   {
      Map<RecordBuffer, Map<Key, HQLHelper>> bufferMap = local.get();
      Map<Key, HQLHelper> cached = bufferMap.get(buffer);
      if (cached == null)
      {
         cached = new HashMap<Key, HQLHelper>();
         bufferMap.put(buffer, cached);
      }
      
      Key key = new Key(where, types, idOnly, sort, joinIface);
      cached.put(key, helper);
   }
   
   /**
    * <code>HQLHelper</code> hashable lookup key.
    */
   private static class Key
   {
      /** HQL where clause, if any */
      private final String where;
      
      /** Hibernate types of query substitution parameters, if any */
      private final Type[] types;
      
      /** Is query a projection query returning primary key ID only? */
      private final boolean idOnly;
      
      /** Order by clause, if any */
      private final String sort;
      
      /** Interface of DMO with which query's records are joined, if any */
      private final Class<?> joinIface;
      
      /**
       * Constructor for a lookup key for the given data.
       *
       * @param   where
       *          Where clause of a query.  May be <code>null</code>.
       * @param   types
       *          Hibernate types of query substitution parameters.  May be an
       *          empty array.
       * @param   idOnly
       *          Indicates whether query is a projection query returning the
       *          primary key ID only.
       * @param   sort
       *          Order by clause of a query.  May be <code>null</code>.
       * @param   joinIface
       *          Interface of DMO with which query's records are joined.  May
       *          be <code>null</code>.
       */
      Key(String where,
          Type[] types,
          boolean idOnly,
          String sort,
          Class<?> joinIface)
      {
         this.where = where;
         this.types = types;
         this.idOnly = idOnly;
         this.sort = sort;
         this.joinIface = joinIface;
      }
      
      /**
       * A hash code implementation consisent with our overridden {@link
       * #equals(Object)} method.
       * 
       * @return  Hash code.
       */
      public int hashCode()
      {
         int result = 17;
         
         result = 37 * result + (where == null ? 0 : where.hashCode());
         result = 37 * result + (idOnly ? 1 : 0);
         result = 37 * result + (sort == null ? 0 : sort.hashCode());
         result = 37 * result + (joinIface == null
                                 ? 0
                                 : joinIface.hashCode());
         for (Type type : types)
         {
            result = 37 * result + type.hashCode();
         }
         
         return result;
      }
      
      /**
       * Test for equivalence with another object.
       * 
       * @param   o
       *          Another <code>ColumnKey</code> instance.
       * 
       * @return  <code>true</code> if <code>o</code> has same internal
       *          state as this object;  else <code>false</code>.
       */
      public boolean equals(Object o)
      {
    	 if(!(o instanceof Key)) return false;
         Key that = (Key) o;
         
         if ((this.joinIface == null || that.joinIface == null) &&
             this.joinIface != that.joinIface)
         {
            return false;
         }
         
         if ((this.where == null || that.where == null) &&
             this.where != that.where)
         {
            return false;
         }
         
         if ((this.sort == null || that.sort == null) &&
             this.sort != that.sort)
         {
            return false;
         }

         boolean same = (this.idOnly == that.idOnly              &&
                         (this.joinIface == null ||
                          this.joinIface.equals(that.joinIface)) &&
                         (this.sort == null ||
                          this.sort.equals(that.sort))           &&
                         (this.where == null ||
                          this.where.equals(that.where)));
         
         int len = this.types.length;
         if (!same || len != that.types.length)
         {
            return false;
         }
         
         for (int i = 0; i < len; i++)
         {
            if (!this.types[i].equals(that.types[i]))
            {
               return false;
            }
         }
         
         return true;
      }
      
      /**
       * Debug method to provide string representation of this object's state.
       * 
       * @return  String representation of this object's state.
       */
      public String toString()
      {
         StringBuilder buf = new StringBuilder();
         buf.append(idOnly ? "id" : "dmo");
         buf.append("@");
         buf.append(where);
         buf.append("@");
         buf.append(sort);
         buf.append("@");
         buf.append(joinIface != null ? joinIface.getName() : "null");
         for (Type type : types)
         {
            buf.append("#");
            buf.append(type);
         }
         
         return buf.toString();
      }
   }
}
