/*
** Module   : FunctionKey.java
** Abstract : Function descriptor which can be used as a hash table key
**
** Copyright (c) 2004-2008, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -T- --JPRM-- ----------------Description-----------------
** 001 ECF 20080312 ADD  @37493  Created initial version. Function descriptor
**                               which can be used as a hash table key.
** 002 ECF 20080508 CHG  @38486  Use Database object rather than database name
**                               String. This allows differentiation of the
**                               same function across multiple databases with
**                               the same name.
*/

package com.goldencode.p2j.persist.hql;

import com.goldencode.p2j.persist.*;

/**
 * Descriptor of a user-defined function, suitable for use as a key in a hash
 * table.
 */
public final class FunctionKey
{
   /** Database name */
   private final Database database;
   
   /** Function name */
   private final String function;
   
   /** Function parameter signature as HQL types */
   private final HQLTypes[] signature;
   
   /**
    * Constructor.
    * 
    * @param   function
    *          Function name.
    * @param   signature
    *          Function parameter signature as HQL types.
    */
   public FunctionKey(String function, HQLTypes[] signature)
   {
      this(null, function, signature);
   }
   
   /**
    * Constructor.
    * 
    * @param   database
    *          Database.
    * @param   function
    *          Function name.
    * @param   signature
    *          Function parameter signature as HQL types.
    */
   public FunctionKey(Database database,
                      String function,
                      HQLTypes[] signature)
   {
      this.database = database;
      this.function = function;
      this.signature = signature;
   }
   
   /**
    * Return hash code for this object.
    * 
    * @return  Hash code.
    * 
    * @see java.lang.Object#hashCode()
    */
   public int hashCode()
   {
      int result = 17;
      
      result = 37 * result + (database != null ? database.hashCode() : 0);
      result = 37 * result + function.hashCode();
      for (HQLTypes type : signature)
      {
         result = 37 * result + type.hashCode();
      }
      
      return result;
   }
   
   /**
    * Test this object for equality with the given object, which is assumed to
    * be an instance of this class.
    * 
    * @return  <code>true</code> if objects are equal, else
    *          <code>false</code>.
    * 
    * @see java.lang.Object#equals(java.lang.Object)
    */
   public boolean equals(Object o)
   {
	  if(!(o instanceof FunctionKey)) return false;
      FunctionKey that = (FunctionKey) o;
      
      boolean bothNull = false;
      
      if (database == null || that.database == null)
      {
         bothNull = (database == that.database);
         if (!bothNull)
         {
            return false;
         }
      }
      
      if (!bothNull && !database.equals(that.database))
      {
         return false;
      }
      
      if (!function.equals(that.function))
      {
         return false;
      }
      
      int len1 = signature.length;
      int len2 = that.signature.length;
      
      if (len1 != len2)
      {
         return false;
      }
      
      for (int i = 0; i < len1; i++)
      {
         if (!signature[i].equals(that.signature[i]))
         {
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Get a string representation of this object's state.
    * 
    * @return  String representation of this object.
    * 
    * @see java.lang.Object#toString()
    */
   public String toString()
   {
      StringBuilder buf = new StringBuilder(database.toString());
      buf.append("::");
      buf.append(function);
      buf.append("(");
      for (int i = 0; i < signature.length; i++)
      {
         if (i > 0)
         {
            buf.append(", ");
         }
         
         buf.append(signature[i]);
      }
      buf.append(")");
      
      return buf.toString();
   }
}