/*
** Module   : LogHelper.java
** Abstract : logging helpers
**
** Copyright (c) 2006-2010, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20060303   @24900 Created initial version, providing helper
**                           methods to initialize the logging 
**                           infrastructure, obtain logger instances,
**                           and generate/format log entries.
** 002 ECF 20060314   @25061 More verbose exception logging. Added error
**                           message and exception type to output. Also
**                           process chained exceptions.
** 003 NVS 20060427   @25809 Internal bind() call was unnecessary but
**                           caused some significant overhead when in
**                           a production environment.
** 004 GES 20070111   @31806 Safety code to allow logging even when
**                           authentication fails (and there is no 
**                           security context).
** 005 NVS 20070126   @32008 "logging" node of the directory is searched
**                           as a server level standard search item.
** 006 EVL 20070609   @34005 Adding explicit import of the class
**                           java.util.logging.Formatter to eliminate
**                           conflict with the same class from java.util
**                           package to be able to compile for Java 6
** 007 GES 20070615   @34068 Fixed honoring the "pattern" XML config
**                           value for the default file handler. Don't
**                           create the file or console handler if they
**                           are disabled in the configuration.
** 008 NVS 20090304   @41408 getSecureLogger() method now takes a boolean
**                           than tells the server vs client environment.
**                           On the client, the console handler will be
**                           always set to Level.OFF to avoid permissions
**                           issues in th admin applet.
** 009 GES 20090422   @41909 Converted to standard string formatting.
** 010 GES 20090722   @43338 Added createThreadName() helper.
** 011 NVS 20090826   @43776 Log file name can be customized to include
**                           the server name by inserting the "%s" token into
**                           the filename template.
** 012 ECF 20090827   @43786 Fixed FileHandler issue. A secure FileHandler is
**                           now stored and shared across all loggers,
**                           including anonymous, secure loggers. This
**                           prevents each anonymous logger from logging to a
**                           separate output file.
** 013 GES 20091103   @44307 Added a facility to allow arbitrary configuration
**                           at initialize time (instead of in the directory).
** 014 SVL 20100319   @44763 Prevented early garbage collection of loggers. 
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.lang.reflect.*;
import java.text.*;
import java.util.*;
import java.util.logging.*;
import java.util.logging.Formatter;

import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.SecurityManager;

/**
 * Provides helpers to initialize the logging infrastructure, obtain logger
 * instances and generate/format log entries.
 */
public class LogHelper
{
   /** Global logging level (all unconfigured handlers inherit this). */
   private static Level global = Level.WARNING; 
   
   /** Logging level for the console handler (<code>OFF</code> to disable). */
   private static Level console = Level.WARNING; 
   
   /** Logging level for the file handler (<code>OFF</code> to disable). */
   private static Level file = Level.OFF;
   
   /** Filename pattern for the file handler. */
   private static String pattern = "%h/p2j_server_%s_%g.log";
   
   /** Maximum file size in megabytes for the file handler. */
   private static int limit = 8;
   
   /** Number of log generations for the file handler. */
   private static int count = 4;
   
   /** Append mode for the file handler. */
   private static boolean append = false;
   
   /** List of logger names and the associated log levels. */
   private static Map<String, Level> loggerLevels = null;
   
   /** Shared file handler */
   private static FileHandler fileHandler = null;

   /**
    * The set of loggers defined into the directory configuration file. This
    * set just keeps hard links to them. Without it these loggers can be GC'd
    * before any child loggers are created (because runtime uses
    * weak-reference structures for saving child loggers) and corresponding
    * logging levels will be lost.  
    */
   private static Set<Logger> definedLoggers = null;
   
   /**
    * Initialize the logging infrastructure using the given configuration.
    *
    * @param    fhClass
    *           Class of a custom file handler instance to use for all loggers.
    *           If this parameter is <code>null</code>, a default handler will
    *           be created.  The class must be public and have a public
    *           constructor of the signature:
    *           <pre>
    *              {ClassName}(String pattern,
    *                          int limit,
    *                          int count,
    *                          boolean append,
    *                          Level logLevel,
    *                          Formatter formatter)
    *           </pre>
    * @param    global
    *           Global logging level (all unconfigured handlers inherit this).
    *           <code>null</code> to use the default.
    * @param    console
    *           Logging level for the console handler (<code>OFF</code> to
    *           disable). <code>null</code> to use the default.
    * @param    file
    *           Logging level for the file handler (<code>OFF</code> to
    *           disable). <code>null</code> to use the default.
    * @param    pattern
    *           Filename pattern for the file handler. <code>null</code> to use
    *           the default.
    * @param    limit
    *           Maximum file size in megabytes for the file handler. Use any
    *           non-positive number to force the default value.
    * @param    count
    *           Number of log generations for the file handler. Use any
    *           non-positive number to force the default value.
    * @param    append
    *           Append mode for the file handler.
    * @param    loglevels
    *           List of logger names and the associated log levels.
    *           
    * @return   Instance of <code>fhClass</code> or <code>null</code> if this
    *           is not the first time initialize has been called.
    */
   public static FileHandler initialize(Class<? extends FileHandler> fhClass,
                                        Level                        global,
                                        Level                        console,
                                        Level                        file,
                                        String                       pattern,
                                        int                          limit,
                                        int                          count,
                                        boolean                      append,
                                        Map<String, Level>           loglevels)
   {
      synchronized (LogHelper.class)
      {
         if (fileHandler != null)
            return null;
         
         if (global != null)
            LogHelper.global = global;
         
         if (console != null)
            LogHelper.console = console;
         
         if (file != null)
            LogHelper.file = file;
         
         if (pattern != null)
            LogHelper.pattern = pattern;
         
         if (limit > 0)
            LogHelper.limit = limit;
         
         if (count > 0)
            LogHelper.count = 0;
         
         LogHelper.append = append;
         
         if (loglevels != null)
            LogHelper.loggerLevels = loglevels;
         
         return initWorker(fhClass);
      }
   }
   
   /**
    * Initialize the logging infrastructure using the configuration stored
    * in the directory.
    *
    * @param    serverName
    *           The name of this server.
    */
   public static void initialize(String serverName)
   {
      initialize(serverName, null);
   }
   
   /**
    * Initialize the logging infrastructure using the configuration stored
    * in the directory.
    *
    * @param    serverName
    *           The name of this server.
    * @param    fhClass
    *           Class of a custom file handler instance to use for all loggers.
    *           If this parameter is <code>null</code>, a default handler will
    *           be created.  The class must be public and have a public
    *           constructor of the signature:
    *           <pre>
    *              {ClassName}(String pattern,
    *                          int limit,
    *                          int count,
    *                          boolean append,
    *                          Level logLevel,
    *                          Formatter formatter)
    *           </pre>
    *           
    * @return   Instance of <code>fhClass</code> or <code>null</code> if this
    *           is not the first time initialize has been called.
    */
   public static FileHandler initialize(String serverName,
                                        Class<? extends FileHandler> fhClass)
   {
      synchronized (LogHelper.class)
      {
         if (fileHandler != null)
            return null;
         
         // read our configuration from the directory
         readConfiguration();
         
         // substitute the server name in the log file name
         StringBuilder sb = new StringBuilder();
   
         int pos = pattern.indexOf("%s");
         if ( pos >= 0)
         {
            if (pos > 0)
            {
               sb.append(pattern.substring(0, pos));
            }
            
            String name = serverName == null ? "" : serverName;
            sb.append(name);
            sb.append(pattern.substring(pos + 2));
            pattern = sb.toString();
         }
         
         return initWorker(fhClass);
      }
   }
      
   /**
    * Retrieve a logging object from the <code>LogManager</code>.
    * The logger retrieved is scoped to the package name of the specified
    * class;  thus, classes in the same package which use this method to
    * retrieve loggers will be sharing the same <code>Log</code> instance.
    *
    * @param   clazz
    *          Class object whose package name is used to scope the Logger.
    *          If <code>null</code> then an anonymous logger (one that has no
    *          name and is not accessible via the namespace) will be 
    *          returned.
    *
    * @return  Logging object for the desired package.
    */
   public static Logger getLogger(Class<?> clazz)
   {
      String name = null;
      
      if (clazz != null)
      {
         name = clazz.getName();         
         name = name.substring(0, name.lastIndexOf("."));
      }
      
      return getLogger(name);
   }
   
   /**
    * Retrieve a logging object from the <code>LogManager</code>.
    * The logger retrieved is scoped to the arbitrary name specified. This
    * may be a classname, a package name or any arbitrary string.  Any name
    * based on a package or class will naturally fit into the hierarchical
    * namespace for all loggers.
    *
    * @param   name
    *          Name used to scope the Logger. If <code>null</code> then an
    *          anonymous logger (one that has no name and is not accessible 
    *          via the namespace) will be returned.
    *
    * @return  Logging object for the desired package.
    */
   public static Logger getLogger(String name)
   {
      Logger log =  name == null ? Logger.getAnonymousLogger()
                                 : Logger.getLogger(name);
                                 
      return log;
   }
   
   /**
    * Retrieve an anonymous logger (one that has no name and is not 
    * accessible via the namespace) from the <code>LogManager</code> and
    * configure it with independent (and inaccessible) instances of the
    * handlers with which the named loggers are configured.
    * <p>
    * When called from the applet client, logging to the console causes
    * the Java Security Manager to throw an Acess Denied exception. To prevent
    * this, the client automatically initializes with the console handler
    * turned off.
    *
    * @param   server
    *          <code>true</code> if called from the server environment,
    *          meaning the console handler will be set to OFF for the client
    *
    * @return  Logging object for the desired package.
    */
   public static Logger getSecureLogger(boolean server)
   {
      // get a logger that isn't in the namespace
      Logger log = Logger.getAnonymousLogger();
      
      // initialize the standard handlers (independent/inaccessible instances)
      if (!server)
      {
         console = Level.OFF;
      }
      
      initLogger(log);

      // ensure that the output is not passed to the parents
      log.setUseParentHandlers(false);
      
      return log;
   }
   
   /**
    * Generate a standard prefix for this string which includes identifiers
    * of the current session's context.
    *
    * @param    msg
    *           The text to append to the session context ids.
    *
    * @return   The standardized log message.
    */
   public static String generate(String msg)
   {
      StringBuffer sb = new StringBuffer(describeContext()).append(msg);
      return sb.toString();
   }
   
   /**
    * Substitute the given parameters into the specification string and then
    * append the result to the standard prefix which includes identifiers
    * of the current session's context.
    *
    * @param    spec
    *           The <code>String.format</code> specification string.
    * @param    parms
    *           The substitution parameters.
    *
    * @return   The standardized log message.
    */
   public static String generate(String spec, Object[] parms)
   {
      return generate(String.format(spec, parms));
   }
   
   /**
    * Substitute the given parameter into the specification string and then
    * append the result to the standard prefix which includes identifiers
    * of the current session's context.
    *
    * @param    spec
    *           The sprintf specification string.
    * @param    obj
    *           The single substitution parameter.
    *
    * @return   The standardized log message.
    */
   public static String generate(String spec, Object obj)
   {
      return generate(spec, new Object[] { obj }); 
   }
   
   /**
    * Substitute the given parameters into the specification string and then
    * append the result to the standard prefix which includes identifiers
    * of the current session's context.
    *
    * @param    spec
    *           The sprintf specification string.
    * @param    o1
    *           The first substitution parameter.
    * @param    o2
    *           The second substitution parameter.
    *
    * @return   The standardized log message.
    */
   public static String generate(String spec, Object o1, Object o2)
   {
      return generate(spec, new Object[] { o1, o2 });
   }
   
   /**
    * Substitute the given parameters into the specification string and then
    * append the result to the standard prefix which includes identifiers
    * of the current session's context.
    *
    * @param    spec
    *           The sprintf specification string.
    * @param    o1
    *           The first substitution parameter.
    * @param    o2
    *           The second substitution parameter.
    * @param    o3
    *           The third substitution parameter.
    *
    * @return   The standardized log message.
    */
   public static String generate(String spec, Object o1, Object o2, Object o3)
   {
      return generate(spec, new Object[] { o1, o2, o3 });
   }
   
   /**
    * Get a simplified string which describes the current context. This is
    * intended primarily for naming a thread.
    *
    * @param    descr
    *           Description of the thread's purpose.
    *
    * @return   Simple description of context.
    */
   public static String createThreadName(String descr)
   {
      SecurityManager sm  = SecurityManager.getInstance();
      String          txt = null;
      
      if (sm != null && SecurityManager.initialized() && sm.hasContext())
      {
         txt = String.format("%s [%08X:%s]",
                             descr,
                             sm.getSessionId(),
                             sm.getUserId());
      }
      
      return txt;
   }
   
   /**
    * Get a string which uniquely describes the current context.  This is
    * intended primarily for debugging or logging purposes.
    *
    * @return  Description of context.
    */
   public static String describeContext()
   {
      SecurityManager sm = SecurityManager.getInstance();
      
      String   spec  = null;
      Object[] parms = null;
      
      if (sm == null || !SecurityManager.initialized() || !sm.hasContext())
      {
         spec = "{%s} ";
         parms = new Object[]
         {
            Thread.currentThread().getName()
         };
      }
      else
      {
         spec = "{%08X:%08X:%s} ";
         parms = new Object[]
         {
            sm.getSessionId(),
            sm.getThreadId(),
            sm.getUserId()
         };         
      }
      
      return String.format(spec, parms);
   }
   
   /**
    * Perform the core initialization using the previously stored and/or
    * default configuration values.
    *
    * @param    fhClass
    *           Class of a custom file handler instance to use for all loggers.
    *           If this parameter is <code>null</code>, a default handler will
    *           be created.  The class must be public and have a public
    *           constructor of the signature:
    *           <pre>
    *              {ClassName}(String pattern,
    *                          int limit,
    *                          int count,
    *                          boolean append,
    *                          Level logLevel,
    *                          Formatter formatter)
    *           </pre>
    *           
    * @return   Instance of <code>fhClass</code>.
    */
   private static FileHandler initWorker(Class<? extends FileHandler> fhClass)
   {
      // clear out current configuration
      LogManager lm = LogManager.getLogManager();
      lm.reset();
      
      // get the global/root logger and set the level
      Logger root = Logger.getLogger("");
      root.setLevel(global);
      
      // store the custom file handler, if a class was provided
      if (fhClass != null)
      {
         try
         {
            // retrieve the expected constructor
            Constructor<? extends FileHandler> ctor =
               fhClass.getConstructor(String.class,
                                      Integer.TYPE,
                                      Integer.TYPE,
                                      Boolean.TYPE,
                                      Level.class,
                                      Formatter.class);
            
            // instantiate the FileHandler subclass
            fileHandler = ctor.newInstance(pattern,
                                           limit * 1024 * 1024,
                                           count,
                                           append,
                                           file,
                                           new CleanFormatter());
         }
         catch (Exception exc)
         {
            System.err.println("WARNING:  " +
                               fhClass.getName() +
                               " could not be instantiated for logging;" +
                               "  default FileHandler will be used!");
            exc.printStackTrace();
         }
      }
      
      // use the default file handler
      if (fileHandler == null)
      {
         try
         {
            fileHandler = new FileHandler(pattern,
                                          limit * 1024 * 1024,
                                          count,
                                          append);
            fileHandler.setLevel(file);
            fileHandler.setFormatter(new CleanFormatter());
         }
         catch (IOException exc)
         {
            System.err.println(
               "WARNING:  could not instantiate file handler for logging");
            exc.printStackTrace();
         }
      }
      
      // initialize the file and console handlers for the root logger
      initLogger(root);
      
      // set the default levels for each specified logger in the namespace
      if (loggerLevels != null)
      {
         definedLoggers = new HashSet<Logger>();

         for (Map.Entry<String, Level> entry : loggerLevels.entrySet())
         {
            String name = entry.getKey();
            Level level = entry.getValue();
            Logger logger = Logger.getLogger(name);
            logger.setLevel(level);

            // prevent loggers from being GC'd (JRE uses weak references 
            // for holding them)
            definedLoggers.add(logger);
         }
      }
      
      return fileHandler;
   }
   
   /**
    * Initialize the given logger with independent instances of the 
    * file and console handlers as configured by {@link #readConfiguration}.
    *
    * @param    log
    *           The logger to configure.
    */
   private static void initLogger(Logger log)
   {
      if (file != null && file != Level.OFF && fileHandler != null)
      {
         log.addHandler(fileHandler);
      }
      
      if (console != null && console != Level.OFF)
      {
         // setup the console handler
         ConsoleHandler ch = new ConsoleHandler();
         ch.setLevel(console);
         ch.setFormatter(new CleanFormatter());
         log.addHandler(ch);
      }      
   }
   
   /**
    * Override the default configuration in this class with one read from
    * the directory.  Searches the following paths (in this precedence order)
    * for the configuration:
    * <p>
    * <pre>
    * Root Paths
    * -------------------------------------------
    * /server/&lt;server_id&gt;/logging
    * /server/default/logging
    * </pre>
    * <p>
    * If found, the local data members are set from the directory. 
    * <p>
    * The following is read (where root is one of the root paths specified
    * above):
    * <p>
    * <pre>
    * Path                                         Type       Purpose
    * -------------------------------------------  ---------  ----------------
    * root + /level                                string     global log level
    * root + /handlers/console/level               string     console handler log level
    * root + /handlers/file/level                  string     file handler log level
    * root + /handlers/file/pattern                string     file handler filename pattern 
    * root + /handlers/file/limit                  integer    file handler log file size limit in kilobytes
    * root + /handlers/file/count                  integer    file handler number of log file generations
    * root + /handlers/file/append                 boolean    file handler should append to current log file
    * root + /loggers/&lt;logger_name&gt;/level    string     log levels for any arbitrary logger name (in the hierarchical namespace)
    * </pre>
    * <p>
    * Logging levels are specified as string names for as defined in the 
    * <code>java.util.logging.Level</code> class.  They are (from least 
    * inclusive to most inclusive):
    * <p>
    * <pre>
    * OFF
    * SEVERE
    * WARNING
    * INFO
    * CONFIG
    * FINE
    * FINER
    * FINEST
    * ALL
    * </pre>
    * <p>
    * Use OFF to disable any logger or handler.  The handler configurations
    * are global (all loggers output to these handlers).
    */
   private static void readConfiguration() 
   {
      // access the directory service
      DirectoryService ds = DirectoryService.getInstance();
      
      // is the directory service unavailable?
      if (ds == null)
      {
         // done!
         return;
      }
      
      // any failure from here out, triggers an unbind if needed
      try
      {
         String path = Utils.findDirectoryNodePath(ds,
                                                   "logging", 
                                                   "container",
                                                   false);
         if (path == null)
            return;
         
         readGlobalLevel(ds, path);
         readConsoleConfig(ds, path);
         readFileConfig(ds, path);
         readLoggerLevels(ds, path);
      }
      
      catch (Exception exc)
      {
         // just allow silent return
      }
   }
   
   /**
    * Read the global default logging level and if it exists and specifies
    * a valid level, override the default.
    * <p>
    * The following is read:
    * <p>
    * <pre>
    * Path                                         Type
    * -------------------------------------------  -------------------
    * root + /level                                string
    * </pre>
    *
    * @param    ds
    *           A bound instance of the directory service.
    * @param    root
    *           The path to the root of the logging configuration.
    */
   private static void readGlobalLevel(DirectoryService ds, String root)
   {
      StringBuffer sb     = new StringBuffer(root).append("/level");
      Level        global = readLevel(ds, sb.toString());
      
      if (global != null)
      {
         LogHelper.global = global;
      }
   }
   
   /**
    * Read the logging level for the console and if it exists and specifies
    * a valid level, override the default.
    * <p>
    * The following is read:
    * <p>
    * <pre>
    * Path                                         Type
    * -------------------------------------------  -------------------
    * root + /handlers/console/level               string
    * </pre>
    *
    * @param    ds
    *           A bound instance of the directory service.
    * @param    root
    *           The path to the root of the logging configuration.
    */
   private static void readConsoleConfig(DirectoryService ds, String root)
   {
      StringBuffer sb = new StringBuffer(root);
      sb.append("/handlers/console/level");
      
      Level console = readLevel(ds, sb.toString());
      
      if (console != null)
      {
         LogHelper.console = console;
      }
   }
   
   /**
    * Read the configuration values for the file handler and if they exist
    * and specify valid values, override the default.
    * <p>
    * The following is read:
    * <p>
    * <pre>
    * Path                                         Type
    * -------------------------------------------  -------------------
    * root + /handlers/file/level                  string
    * root + /handlers/file/pattern                string
    * root + /handlers/file/limit                  integer
    * root + /handlers/file/count                  integer
    * root + /handlers/file/append                 boolean
    * </pre>
    *
    * @param    ds
    *           A bound instance of the directory service.
    * @param    root
    *           The path to the root of the logging configuration.
    */
   private static void readFileConfig(DirectoryService ds, String root)
   {
      StringBuffer sb = new StringBuffer(root);
      sb.append("/handlers/file");
      
      String nodeClass = ds.getNodeClass(sb.toString());
      
      if (nodeClass != null)
      {
         if (nodeClass.equals("/meta/class/container"))
         {
            // we have a configuration
            
            // try to read the level
            sb = new StringBuffer(root);
            sb.append("/handlers/file/level");
            
            Level file = readLevel(ds, sb.toString());
            
            if (file != null)
            {
               LogHelper.file = file;
            }
            
            // try to read the filename pattern
            sb = new StringBuffer(root);
            sb.append("/handlers/file/pattern");
            
            String pattern = ds.getNodeString(sb.toString(), "value");
            
            if (pattern != null)
            {
               LogHelper.pattern = pattern;
            }
            
            // try to read the limit
            sb = new StringBuffer(root);
            sb.append("/handlers/file/limit");
            
            Integer limit = ds.getNodeInteger(sb.toString(), "value");
            
            if (limit != null)
            {
               LogHelper.limit = limit.intValue();
            }
            
            // try to read the count
            sb = new StringBuffer(root);
            sb.append("/handlers/file/count");
            
            Integer count = ds.getNodeInteger(sb.toString(), "value");
            
            if (count != null)
            {
               LogHelper.count = count.intValue();
            }
            
            // try to read the append flag
            sb = new StringBuffer(root);
            sb.append("/handlers/file/append");
            
            Boolean append = ds.getNodeBoolean(sb.toString(), "value");
            
            if (append != null)
            {
               LogHelper.append = append.booleanValue();
            }            
         }
      }            
   }
   
   /**
    * Read the log level for an arbitrary list of loggers (as specified by
    * their hierarchical names).
    * <p>
    * The following is read (may be 0 or more of these entries, each name
    * must be unique):
    * <p>
    * <pre>
    * Path                                         Type
    * -------------------------------------------  -------------------
    * root + /loggers/&lt;logger_name&gt;/level    string
    * </pre>
    *
    * @param    ds
    *           A bound instance of the directory service.
    * @param    root
    *           The path to the root of the logging configuration.
    */
   private static void readLoggerLevels(DirectoryService ds, String root)
   {
      StringBuffer sb = new StringBuffer(root);
      sb.append("/loggers");
      
      String base = sb.toString();
      
      String nodeClass = ds.getNodeClass(base);
      
      if (nodeClass != null)
      {
         if (nodeClass.equals("/meta/class/container"))
         {
            // make a list of child nodes
            String[] list = ds.enumerateNodes(base);
            
            for (int i = 0; i < list.length; i++)
            {
               // build the full path
               sb = new StringBuffer(base).append("/").append(list[i]);
               
               // try to read the level
               Level lvl = readLevel(ds, sb.toString());
               
               // save the result
               if (lvl != null)
               {
                  if (loggerLevels == null)
                  {
                     loggerLevels = new LinkedHashMap<String, Level>();
                  }
                  
                  loggerLevels.put(list[i], lvl);
               }               
            }
         }
      }            
   }
   
   /**
    * Read the given node as a string and convert the string name into a
    * <code>Level</code> object.
    *
    * @param    ds
    *           A bound instance of the directory service.
    * @param    nodeId
    *           The node to read.
    *
    * @return   The level read or <code>null</code> on any failure.
    */
   private static Level readLevel(DirectoryService ds, String nodeId)
   {
      Level level = null;
      String nodeClass = ds.getNodeClass(nodeId);
      
      if (nodeClass != null)
      {
         if (nodeClass.equals("/meta/class/string"))
         {
            String lvl = ds.getNodeString(nodeId, "value");
            
            if (lvl != null)
            {
               try
               {
                  level = Level.parse(lvl);
               }
               
               catch (IllegalArgumentException iae)
               {
                  // invalid level name, return null
               }
            }
         }
      }      
      
      return level;
   }
   
   /**
    * Simple formatter to emit a very clean, condensed output.
    */
   public static class CleanFormatter
   extends Formatter
   {
      /** Date format string. */
      private static final String FMT = "MM/dd/yyyy HH:mm:ss z";
      
      /** Date and time formatter. */
      private static final ThreadLocal<SimpleDateFormat> sdf = new ThreadLocal<SimpleDateFormat>() {
          @Override 
          protected SimpleDateFormat initialValue() {
              return new SimpleDateFormat("FMT");
          }
      };
      
      /** Platform-specific line separator. */
      private static final String SEP = System.getProperty("line.separator");
      
      /**
       * Format the given log record and return the formatted string.
       *
       * @param    record
       *           The log record to be formatted. 
       *
       * @return   The formatted log record.
       */
      public String format(LogRecord record)
      {
         String spec = "[%s] (%s:%s) %s%s";
         
         Date dat = new Date(record.getMillis());
         
         Object[] parms = new Object[]
         {
            sdf.get().format(dat),
            record.getSourceClassName(),
            record.getLevel().toString(),
            record.getMessage(),
            SEP
         };
         
         String out = String.format(spec, parms);
         
         Throwable t = record.getThrown();
         
         if (t != null)
         {
            StringBuffer sb = new StringBuffer(out);
            dumpThrowable(t, sb);
            out = sb.toString();
         }
         
         return out;
      }
      
      /**
       * Dump all details of the throwable into the given string buffer.
       * Includes exception type, message, complete stack trace, and chained
       * throwables, if any.
       *
       * @param    t
       *           The throwable to dump.
       * @param    buf
       *           String buffer into which details are written.
       */
      private void dumpThrowable(Throwable t, StringBuffer buf)
      {
         Throwable cause = t;
         do
         {
            if (cause != t)
            {
               buf.append("Caused by: ");
            }
            
            buf.append(cause.getClass().getName());
            String msg = cause.getMessage();
            if (msg != null && msg.trim().length() > 0)
            {
               buf.append(": ");
               buf.append(msg);
            }
            buf.append(SEP);
            
            StackTraceElement[] trace = cause.getStackTrace();
            
            for (int i = 0; i < trace.length; i++)
            {
               buf.append("        at ");
               buf.append(trace[i]);
               buf.append(SEP);
            }
            
            cause = cause.getCause();
         }
         while (cause != null);
      }
   }   
}
